@@ -1,9 +1,8 @@
-# -*- encoding: utf-8 -*-
 """
 
     Definitions used for tokenization of Icelandic text
 
-    Copyright (C) 2020 Miðeind ehf.
+    Copyright (C) 2021 Miðeind ehf.
     Original author: Vilhjálmur Þorsteinsson
 
     This software is licensed under the MIT License:
@@ -29,45 +28,81 @@
 
 """
 
-from __future__ import absolute_import
-from __future__ import unicode_literals
+from typing import (
+    Dict,
+    FrozenSet,
+    Tuple,
+    Union,
+    Callable,
+    List,
+    Sequence,
+    Optional,
+    NamedTuple,
+    cast,
+)
 
-from typing import Dict, Tuple, Union, Callable
-
-import sys
 import re
 
 
-# Mask away difference between Python 2 and 3
-if sys.version_info >= (3, 0):
-    items = lambda d: d.items()
-    keys = lambda d: d.keys()
-    make_str = lambda s: s
-    unicode_chr = lambda c: chr(c)
-    is_str = lambda s: isinstance(s, str)
-    if sys.version_info >= (3, 5):
-        # On Python >= 3.5, the typing module is available
-        from typing import Callable
-else:
-    items = lambda d: d.iteritems()
-    keys = lambda d: d.iterkeys()
+BeginTuple = Tuple[int, Optional[int]]
+PunctuationTuple = Tuple[int, str]
+NumberTuple = Tuple[float, Optional[List[str]], Optional[List[str]]]
+DateTimeTuple = Tuple[int, int, int]
+MeasurementTuple = Tuple[str, float]
+TimeStampTuple = Tuple[int, int, int, int, int, int]
+AmountTuple = Tuple[float, str, Optional[List[str]], Optional[List[str]]]
+TelnoTuple = Tuple[str, str]
+CurrencyTuple = Tuple[str, Optional[List[str]], Optional[List[str]]]
 
-    # pylint: disable=undefined-variable
-    def make_str(s):
-        if isinstance(s, unicode):
-            return s
-        # Assume that incoming byte strings are UTF-8 encoded
-        return s.decode("utf-8")
+BIN_Tuple = NamedTuple(
+    "BIN_Tuple",
+    [
+        ("stofn", str),
+        ("utg", int),
+        ("ordfl", str),
+        ("fl", str),
+        ("ordmynd", str),
+        ("beyging", str),
+    ],
+)
+BIN_TupleList = Sequence[BIN_Tuple]
 
-    unicode_chr = lambda c: unichr(c)
-    is_str = lambda s: isinstance(s, (unicode, str))
+PersonNameTuple = NamedTuple(
+    "PersonNameTuple",
+    [
+        ("name", str),
+        ("gender", Optional[str]),
+        ("case", Optional[str]),
+    ]
+)
+PersonNameList = Sequence[PersonNameTuple]
 
+# All possible contents of the Tok.val attribute
+ValType = Union[
+    None,
+    int,  # YEAR, ORDINAL
+    str,  # USERNAME
+    BeginTuple,  # S_BEGIN
+    PunctuationTuple,  # PUNCTUATION, NUMWLETTER
+    MeasurementTuple,  # MEASUREMENT
+    TelnoTuple,  # TELNO
+    DateTimeTuple,  # DATE, TIME
+    TimeStampTuple,  # TIMESTAMP
+    NumberTuple,  # PERCENT, NUMBER
+    AmountTuple,  # AMOUNT
+    CurrencyTuple,  # CURRENCY
+    BIN_TupleList,  # WORD
+    PersonNameList,  # PERSON
+]
 
-ACCENT = unicode_chr(769)
-UMLAUT = unicode_chr(776)
-SOFT_HYPHEN = unicode_chr(173)
-ZEROWIDTH_SPACE = unicode_chr(8203)
-ZEROWIDTH_NBSP = unicode_chr(65279)
+# This seems to be needed as a workaround for Pylance/Pyright
+_escape = cast(Callable[[str], str], re.escape)
+
+ACCENT = chr(769)
+UMLAUT = chr(776)
+SOFT_HYPHEN = chr(173)
+ZEROWIDTH_SPACE = chr(8203)
+ZEROWIDTH_NBSP = chr(65279)
 
 # Preprocessing of unicode characters before tokenization
 UNICODE_REPLACEMENTS = {
@@ -101,9 +136,12 @@ UNICODE_REPLACEMENTS = {
     ZEROWIDTH_NBSP: "",
 }
 UNICODE_REGEX = re.compile(
-    r"|".join(map(re.escape, keys(UNICODE_REPLACEMENTS))), re.UNICODE
+    r"|".join(map(_escape, UNICODE_REPLACEMENTS.keys())), re.UNICODE
 )
 
+# Used for the first step of token splitting
+ROUGH_TOKEN_REGEX = re.compile(r"(\s*)([^\s]*)", re.UNICODE)
+
 # Hyphens are normalized to '-'
 HYPHEN = "-"  # Normal hyphen
 EN_DASH = "\u2013"  # "–"
@@ -172,9 +210,7 @@ CLOCK_ABBREVS = frozenset(("kl", "kl.", "klukkan"))
 TELNO_PREFIXES = "45678"
 
 # Known telephone country codes
-COUNTRY_CODES = frozenset((
-    "354", "+354", "00354",
-))
+COUNTRY_CODES = frozenset(("354", "+354", "00354",))
 
 # Words that can precede a year number; will be assimilated into the year token
 YEAR_WORD = frozenset(("árið", "ársins", "árinu"))
@@ -386,7 +422,7 @@ CURRENCY_SYMBOLS = {
 SINGLECHAR_FRACTIONS = "↉⅒⅑⅛⅐⅙⅕¼⅓½⅖⅔⅜⅗¾⅘⅝⅚⅞"
 
 # Derived unit : (base SI unit, conversion factor/function)
-SI_UNITS = {
+SI_UNITS: Dict[str, Tuple[str, Union[float, Callable[[float], float]]]] = {
     # Distance
     "m": ("m", 1.0),
     "mm": ("m", 1.0e-3),
@@ -447,7 +483,7 @@ SI_UNITS = {
     "MWh": ("J", 3.6e9),
     "kWst": ("J", 3.6e6),
     "MWst": ("J", 3.6e9),
-    "kcal": ("J", 4184),
+    "kcal": ("J", 4184.0),
     "cal": ("J", 4.184),
     # Power
     "W": ("W", 1.0),
@@ -476,21 +512,29 @@ SI_UNITS = {
     # Percentage and promille
     "%": ("%", 1.0),
     "‰": ("‰", 0.1),
-}  # type: Dict[str, Tuple[str, Union[float, Callable[[float], float]]]]
+    # Velocity
+    "m/s": ("m/s", 1.0),
+    "km/klst": ("m/s", 1000.0 / (60 * 60)),
+    # "km/klst.": ("m/s", 1000.0/(60*60)),
+}
 
 DIRECTIONS = {
     "N": "Norður",
 }
 
-SI_UNITS_SET = frozenset(keys(SI_UNITS))
+_unit_lambda: Callable[[str], str] = lambda unit: unit + r"(?!\w)" if unit[
+    -1
+].isalpha() else unit
+
+SI_UNITS_SET: FrozenSet[str] = frozenset(SI_UNITS.keys())
 SI_UNITS_REGEX_STRING = r"|".join(
     map(
         # If the unit ends with a letter, don't allow the next character
         # after it to be a letter (i.e. don't match '220Volts' as '220V')
-        lambda unit: unit + r"(?!\w)" if unit[-1].isalpha() else unit,
+        _unit_lambda,
         # Sort in descending order by length, so that longer strings
         # are matched before shorter ones
-        sorted(keys(SI_UNITS), key=lambda s: len(s), reverse=True)
+        sorted(SI_UNITS.keys(), key=len, reverse=True),
     )
 )
 SI_UNITS_REGEX = re.compile(r"({0})".format(SI_UNITS_REGEX_STRING), re.UNICODE)
@@ -499,8 +543,8 @@ CURRENCY_REGEX_STRING = r"|".join(
     map(
         # Sort in descending order by length, so that longer strings
         # are matched before shorter ones
-        re.escape,
-        sorted(keys(CURRENCY_SYMBOLS), key=lambda s: len(s), reverse=True)
+        _escape,
+        sorted(CURRENCY_SYMBOLS.keys(), key=lambda s: len(s), reverse=True),
     )
 )
 
@@ -509,21 +553,18 @@ UNIT_REGEX_STRING = SI_UNITS_REGEX_STRING + r"|" + CURRENCY_REGEX_STRING
 
 # Icelandic-style number, followed by a unit
 NUM_WITH_UNIT_REGEX1 = re.compile(
-    r"([-+]?\d+(\.\d\d\d)*(,\d+)?)({0})".format(UNIT_REGEX_STRING),
-    re.UNICODE
+    r"([-+]?\d+(\.\d\d\d)*(,\d+)?)({0})".format(UNIT_REGEX_STRING), re.UNICODE
 )
 
 # English-style number, followed by a unit
 NUM_WITH_UNIT_REGEX2 = re.compile(
-    r"([-+]?\d+(,\d\d\d)*(\.\d+)?)({0})".format(UNIT_REGEX_STRING),
-    re.UNICODE
+    r"([-+]?\d+(,\d\d\d)*(\.\d+)?)({0})".format(UNIT_REGEX_STRING), re.UNICODE
 )
 
 # One or more digits, followed by a unicode vulgar fraction char (e.g. '2½')
 # and a unit (SI, percent or currency symbol)
 NUM_WITH_UNIT_REGEX3 = re.compile(
-    r"(\d+)([\u00BC-\u00BE\u2150-\u215E])({0})".format(UNIT_REGEX_STRING),
-    re.UNICODE
+    r"(\d+)([\u00BC-\u00BE\u2150-\u215E])({0})".format(UNIT_REGEX_STRING), re.UNICODE
 )
 
 
@@ -574,7 +615,7 @@ ORDINAL_NUMBERS = {
     "4ðu": 4,
     "5ti": 5,
     "5ta": 5,
-    "5tu": 5
+    "5tu": 5,
 }
 
 # Handling of Roman numerals
@@ -591,7 +632,7 @@ ROMAN_NUMERAL_MAP = tuple(
 )
 
 
-def roman_to_int(s):
+def roman_to_int(s: str) -> int:
     """ Quick and dirty conversion of an already validated Roman numeral to integer """
     # Adapted from http://code.activestate.com/recipes/81611-roman-numerals/
     i = result = 0
@@ -705,7 +746,15 @@ ISK_AMOUNT_PRECEDING = frozenset(("kr.", "kr", "krónur"))
 
 # URL prefixes. Note that this list should not contain www since
 # www.something.com is a domain token, not a URL token.
-URL_PREFIXES = ("http://", "https://", "file://")
+URL_PREFIXES = (
+    "http://",
+    "https://",
+    "file://",
+    "ftp://",
+    "ssh://",
+    "sftp://",
+    "smb://",
+)
 
 TOP_LEVEL_DOMAINS = frozenset(
     (
@@ -743,6 +792,7 @@ TOP_LEVEL_DOMAINS = frozenset(
         "store",
         "gdn",
         "art",
+        "events",
         # ccTLDs
         "ac",
         "ad",
@@ -862,7 +912,7 @@ TOP_LEVEL_DOMAINS = frozenset(
         "km",
         "kn",
         "kp",
-        # "kr", # Gives us trouble with "kr" abbreviation (e.g. "þús.kr" is a legitimate domain name)
+        # "kr", # Clashes with "kr" abbreviation (e.g. "þús.kr" is a legitimate domain name)
         "kw",
         "ky",
         "kz",
@@ -991,9 +1041,6 @@ TOP_LEVEL_DOMAINS = frozenset(
     )
 )
 
-# This is a small hack to satisfy the Mypy type checker
-_escape = lambda s: re.escape(s)  # type: Callable[[str], str]
-
 # Regex to recognise domain names
 MIN_DOMAIN_LENGTH = 4  # E.g. "t.co"
 DOMAIN_REGEX = re.compile(
@@ -1141,7 +1188,7 @@ MOLECULE_FILTER = re.compile(r"\d")
 KT_MAGIC = [3, 2, 7, 6, 5, 4, 0, 3, 2]
 
 
-def valid_ssn(kt):
+def valid_ssn(kt: str) -> bool:
     """ Validate Icelandic social security number """
     if not kt or len(kt) != 11 or kt[6] != "-":
         return False
@@ -1179,7 +1226,7 @@ HTML_ESCAPES = {
     "amp": "&",
     "lt": "<",
     "gt": ">",
-    "quot": "\"",
+    "quot": '"',
     "apos": "'",
     "bdquo": "„",
     "ldquo": "“",
@@ -1225,6 +1272,7 @@ HTML_ESCAPES = {
 
 ESCAPES_REGEX = r"|".join(HTML_ESCAPES.keys())
 HTML_ESCAPE_REGEX = re.compile(
-    r"&((#x[0-9a-fA-F]{r1})|(#\d{r2})|({ex}))\;"
-    .format(r1="{1,8}", r2="{1,10}", ex=ESCAPES_REGEX)
+    r"&((#x[0-9a-fA-F]{r1})|(#\d{r2})|({ex}))\;".format(
+        r1="{1,8}", r2="{1,10}", ex=ESCAPES_REGEX
+    )
 )
